@@ -1,6 +1,4 @@
-#![allow(clippy::uninlined_format_args)]
-
-use fish_build_helper::{fish_build_dir, workspace_root};
+use fish_build_helper::{env_var, fish_build_dir, workspace_root};
 use rsconf::Target;
 use std::env;
 use std::path::{Path, PathBuf};
@@ -30,16 +28,17 @@ fn main() {
     );
 
     // Some build info
-    rsconf::set_env_value("BUILD_TARGET_TRIPLE", &env::var("TARGET").unwrap());
-    rsconf::set_env_value("BUILD_HOST_TRIPLE", &env::var("HOST").unwrap());
-    rsconf::set_env_value("BUILD_PROFILE", &env::var("PROFILE").unwrap());
+    rsconf::set_env_value("BUILD_TARGET_TRIPLE", &env_var("TARGET").unwrap());
+    rsconf::set_env_value("BUILD_HOST_TRIPLE", &env_var("HOST").unwrap());
+    rsconf::set_env_value("BUILD_PROFILE", &env_var("PROFILE").unwrap());
 
     let version = &get_version(&env::current_dir().unwrap());
     // Per https://doc.rust-lang.org/cargo/reference/build-scripts.html#inputs-to-the-build-script,
     // the source directory is the current working directory of the build script
     rsconf::set_env_value("FISH_BUILD_VERSION", version);
 
-    std::env::set_var("FISH_BUILD_VERSION", version);
+    // safety: single-threaded code.
+    unsafe { std::env::set_var("FISH_BUILD_VERSION", version) };
 
     // These are necessary if built with embedded functions,
     // but only in release builds (because rust-embed in debug builds reads from the filesystem).
@@ -50,9 +49,6 @@ fn main() {
     #[cfg(feature = "gettext-extract")]
     rsconf::rebuild_if_env_changed("FISH_GETTEXT_EXTRACTION_FILE");
 
-    rsconf::rebuild_if_path_changed("src/libc.c");
-    cc::Build::new().file("src/libc.c").compile("flibc.a");
-
     let build = cc::Build::new();
     let mut target = Target::new_from(build).unwrap();
     // Keep verbose mode on until we've ironed out rust build script stuff
@@ -60,7 +56,9 @@ fn main() {
     detect_cfgs(&mut target);
 
     #[cfg(all(target_env = "gnu", target_feature = "crt-static"))]
-    compile_error!("Statically linking against glibc has unavoidable crashes and is unsupported. Use dynamic linking or link statically against musl.");
+    compile_error!(
+        "Statically linking against glibc has unavoidable crashes and is unsupported. Use dynamic linking or link statically against musl."
+    );
 }
 
 /// Check target system support for certain functionality dynamically when the build is invoked,
@@ -83,6 +81,7 @@ fn detect_cfgs(target: &mut Target) {
         ("", &(|_: &Target| false) as &dyn Fn(&Target) -> bool),
         ("apple", &detect_apple),
         ("bsd", &detect_bsd),
+        ("using_cmake", &|_| option_env!("FISH_CMAKE_BINARY_DIR").is_some()),
         ("cygwin", &detect_cygwin),
         ("small_main_stack", &has_small_stack),
         // See if libc supports the thread-safe localeconv_l(3) alternative to localeconv(3).
@@ -117,7 +116,7 @@ fn detect_apple(_: &Target) -> bool {
 
 fn detect_cygwin(_: &Target) -> bool {
     // Cygwin target is usually cross-compiled.
-    std::env::var("CARGO_CFG_TARGET_OS").unwrap() == "cygwin"
+    env_var("CARGO_CFG_TARGET_OS").unwrap() == "cygwin"
 }
 
 /// Detect if we're being compiled for a BSD-derived OS, allowing targeting code conditionally with
@@ -129,11 +128,10 @@ fn detect_cygwin(_: &Target) -> bool {
 fn detect_bsd(_: &Target) -> bool {
     // Instead of using `uname`, we can inspect the TARGET env variable set by Cargo. This lets us
     // support cross-compilation scenarios.
-    let mut target = std::env::var("TARGET").unwrap();
+    let mut target = env_var("TARGET").unwrap();
     if !target.chars().all(|c| c.is_ascii_lowercase()) {
         target = target.to_ascii_lowercase();
     }
-    #[allow(clippy::let_and_return)] // for old clippy
     let is_bsd = target.ends_with("bsd") || target.ends_with("dragonfly");
     #[cfg(any(
         target_os = "dragonfly",
@@ -163,19 +161,16 @@ fn has_small_stack(_: &Target) -> bool {
     {
         use core::ffi;
 
-        extern "C" {
-            fn pthread_get_stacksize_np(thread: *const ffi::c_void) -> usize;
-            fn pthread_self() -> *const ffi::c_void;
+        unsafe extern "C" {
+            unsafe fn pthread_get_stacksize_np(thread: *const ffi::c_void) -> usize;
+            unsafe fn pthread_self() -> *const ffi::c_void;
         }
 
         // build.rs is executed on the main thread, so we are getting the main thread's stack size.
         // Modern macOS versions default to an 8 MiB main stack but legacy OS X have a 0.5 MiB one.
         let stack_size = unsafe { pthread_get_stacksize_np(pthread_self()) };
         const TWO_MIB: usize = 2 * 1024 * 1024 - 1;
-        match stack_size {
-            0..=TWO_MIB => true,
-            _ => false,
-        }
+        stack_size <= TWO_MIB
     }
 }
 
@@ -183,59 +178,77 @@ fn setup_paths() {
     #[cfg(windows)]
     use unix_path::{Path, PathBuf};
 
-    fn get_path(name: &str, default: &str, onvar: &Path) -> PathBuf {
-        let mut var = PathBuf::from(env::var(name).unwrap_or(default.to_string()));
-        if var.is_relative() {
-            var = onvar.join(var);
+    fn overridable_path(
+        env_var_name: &str,
+        f: impl FnOnce(Option<String>) -> Option<PathBuf>,
+    ) -> Option<PathBuf> {
+        rsconf::rebuild_if_env_changed(env_var_name);
+        let maybe_path = f(env_var(env_var_name));
+        if let Some(path) = maybe_path.as_ref() {
+            rsconf::set_env_value(env_var_name, path.to_str().unwrap());
         }
-        var
+        maybe_path
     }
 
-    let prefix = PathBuf::from(env::var("PREFIX").unwrap_or("/usr/local".to_string()));
-    rsconf::rebuild_if_env_changed("PREFIX");
-    rsconf::set_env_value("PREFIX", prefix.to_str().unwrap());
-
-    let datadir = get_path("DATADIR", "share/", &prefix);
-
-    let sysconfdir = get_path(
-        "SYSCONFDIR",
-        // Embedded builds use "/etc," not "./share/etc".
-        if cfg!(feature = "embed-data") {
-            "/etc/"
+    fn join_if_relative(parent_if_relative: &Path, path: String) -> PathBuf {
+        let path = PathBuf::from(path);
+        if path.is_relative() {
+            parent_if_relative.join(path)
         } else {
-            "etc/"
-        },
-        &datadir,
-    );
-    rsconf::set_env_value("SYSCONFDIR", sysconfdir.to_str().unwrap());
-    rsconf::rebuild_if_env_changed("SYSCONFDIR");
-
-    #[cfg(not(feature = "embed-data"))]
-    {
-        rsconf::set_env_value("DATADIR", datadir.to_str().unwrap());
-        rsconf::rebuild_if_env_changed("DATADIR");
-
-        let bindir = get_path("BINDIR", "bin/", &prefix);
-        rsconf::set_env_value("BINDIR", bindir.to_str().unwrap());
-        rsconf::rebuild_if_env_changed("BINDIR");
-
-        let localedir = get_path("LOCALEDIR", "locale/", &datadir);
-        let localedir = localedir.to_str().unwrap();
-        assert!(!localedir.is_empty(), "empty LOCALEDIR is not supported");
-        rsconf::set_env_value("LOCALEDIR", localedir);
-        rsconf::rebuild_if_env_changed("LOCALEDIR");
-
-        let docdir = get_path("DOCDIR", "doc/fish", &datadir);
-        rsconf::set_env_value("DOCDIR", docdir.to_str().unwrap());
-        rsconf::rebuild_if_env_changed("DOCDIR");
+            path
+        }
     }
+
+    let prefix = overridable_path("PREFIX", |env_prefix| {
+        Some(PathBuf::from(
+            env_prefix.unwrap_or("/usr/local".to_string()),
+        ))
+    })
+    .unwrap();
+
+    overridable_path("SYSCONFDIR", |env_sysconfdir| {
+        Some(join_if_relative(
+            &prefix,
+            env_sysconfdir.unwrap_or(
+                // Embedded builds use "/etc," not "$PREFIX/etc".
+                if cfg!(feature = "embed-data") {
+                    "/etc/"
+                } else {
+                    "etc/"
+                }
+                .to_string(),
+            ),
+        ))
+    });
+
+    let default_ok = !cfg!(feature = "embed-data");
+    let datadir = overridable_path("DATADIR", |env_datadir| {
+        let default = default_ok.then_some("share/".to_string());
+        env_datadir
+            .or(default)
+            .map(|p| join_if_relative(&prefix, p))
+    });
+    overridable_path("BINDIR", |env_bindir| {
+        let default = default_ok.then_some("bin/".to_string());
+        env_bindir.or(default).map(|p| join_if_relative(&prefix, p))
+    });
+    overridable_path("DOCDIR", |env_docdir| {
+        let default = default_ok.then_some("doc/fish".to_string());
+        env_docdir.or(default).map(|p| {
+            join_if_relative(
+                &datadir
+                    .expect("Setting DOCDIR without setting DATADIR is not currently supported"),
+                p,
+            )
+        })
+    });
 }
 
 fn get_version(src_dir: &Path) -> String {
     use std::fs::read_to_string;
     use std::process::Command;
 
-    if let Ok(var) = std::env::var("FISH_BUILD_VERSION") {
+    if let Some(var) = env_var("FISH_BUILD_VERSION") {
         return var;
     }
 
