@@ -19,6 +19,7 @@ import subprocess
 import sys
 import tempfile
 import threading
+import typing
 from itertools import chain
 
 COMMON_WSL_CMD_PATHS = (
@@ -68,14 +69,6 @@ def find_executable(exe, paths=()):
             return proposed_path
 
 
-def isMacOS10_12_5_OrLater():
-    """Return whether this system is macOS 10.12.5 or a later version."""
-    try:
-        return [int(x) for x in platform.mac_ver()[0].split(".")] >= [10, 12, 5]
-    except ValueError:
-        return False
-
-
 def is_wsl():
     """Return whether we are running under the Windows Subsystem for Linux"""
     if "linux" in platform.system().lower() and os.access("/proc/version", os.R_OK):
@@ -122,23 +115,12 @@ def is_chromeos_garcon():
 
 
 def run_fish_cmd(text):
-    # Ensure that fish is using UTF-8.
-    ctype = os.environ.get("LC_ALL", os.environ.get("LC_CTYPE", os.environ.get("LANG")))
-    env = None
-    if ctype is None or re.search(r"\.utf-?8$", ctype, flags=re.I) is None:
-        # override LC_CTYPE with en_US.UTF-8
-        # We're assuming this locale exists.
-        # Fish makes the same assumption in config.fish
-        env = os.environ.copy()
-        env.update(LC_CTYPE="en_US.UTF-8", LANG="en_US.UTF-8")
-
     print("$ " + text)
     p = subprocess.Popen(
         [FISH_BIN_PATH],
         stdin=subprocess.PIPE,
         stdout=subprocess.PIPE,
         stderr=subprocess.PIPE,
-        env=env,
     )
     out, err = p.communicate(text.encode("utf-8"))
     out = out.decode("utf-8", "replace")
@@ -229,7 +211,7 @@ def parse_color(color_str):
     background_color = ""
     underline_color = ""
     bold = False
-    underline = False
+    underline = None
     italics = False
     dim = False
     reverse = False
@@ -243,11 +225,13 @@ def parse_color(color_str):
         elif comp == "--underline" or comp == "-u":
             underline = "single"
         elif comp.startswith("--underline="):
-            underline = comp.stripprefix("--underline=")
+            # TODO(python>3.8): use removeprefix
+            underline = comp[len("--underline=") :]
         elif comp.startswith(
             "-u"
         ):  # Multiple short options like "-rucurly" are not yet supported.
-            underline = comp.stripprefix("-u")
+            # TODO(python>3.8): use removeprefix
+            underline = comp[len("-u") :]
         elif comp == "--italics" or comp == "-i":
             italics = True
         elif comp == "--dim" or comp == "-d":
@@ -257,10 +241,14 @@ def parse_color(color_str):
         else:
 
             def parse_opt(
-                current_best: str, i: int, long_opt: str, short_opt: str | None
+                current_best: str, i: int, long_opt: str, short_opt: Optional[str]
             ) -> str:
                 if comp.startswith(long_opt):
                     c = comp[len(long_opt) :]
+                    if c[0] == "=":
+                        # There was a = between the long option and the value.
+                        # i.e. support also --background=red, not just --background red
+                        c = c[1:]
                     parsed_c = parse_one_color(c)
                     # We prefer the unparsed version - if it says "brgreen", we use brgreen,
                     # instead of 00ff00
@@ -1274,13 +1262,13 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
         # In some cases it'll give us the encoding as well,
         # ("application/json;charset=utf-8")
         # but we don't currently care.
-        ctype = self.headers["content-type"].split(";")[0]
+        content_type = self.headers["content-type"].split(";")[0]
 
-        if ctype == "application/x-www-form-urlencoded":
+        if content_type == "application/x-www-form-urlencoded":
             length = int(self.headers["content-length"])
             url_str = self.rfile.read(length).decode("utf-8")
             postvars = parse_qs(url_str, keep_blank_values=1)
-        elif ctype == "application/json":
+        elif content_type == "application/json":
             length = int(self.headers["content-length"])
             # This used to use the provided encoding, but we use utf-8
             # all around the place and nobody has ever complained.
@@ -1290,7 +1278,7 @@ class FishConfigHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
             # If that happens to anyone we expect bug reports.
             url_str = self.rfile.read(length).decode("utf-8")
             postvars = json.loads(url_str)
-        elif ctype == "multipart/form-data":
+        elif content_type == "multipart/form-data":
             # This used to be a thing, as far as I could find there's
             # no use anymore, but let's keep an error around just in case.
             return self.send_error(500)
@@ -1417,7 +1405,7 @@ fish_bin_path = None
 fish_bin_name = "fish.exe" if is_windows() else "fish"
 
 if not fish_bin_dir:
-    print("The $__fish_bin_dir environment variable is not set. " "Looking in $PATH...")
+    print("The $__fish_bin_dir environment variable is not set. Looking in $PATH...")
     fish_bin_path = find_executable(fish_bin_name)
     if not fish_bin_path:
         print("fish could not be found. Is fish installed correctly?")
@@ -1538,8 +1526,9 @@ print("%sHit ENTER to stop.%s" % (ENTER_BOLD_MODE, EXIT_ATTRIBUTE_MODE))
 
 
 def runThing():
-    if isMacOS10_12_5_OrLater():
-        subprocess.check_call(["open", fileurl])
+    if os.environ.get("BROWSER") == "true":
+        # Don't start a browser in this case (see issue #11926)
+        pass
     elif is_wsl():
         cmd_path = find_executable("cmd.exe", COMMON_WSL_CMD_PATHS)
         if cmd_path:
@@ -1600,7 +1589,15 @@ def capture_enter(port):
 
 def get_windows_signal():
     """Using socket as a replacement for stdin on Windows."""
-    (sig, sig_port) = create_socket(8000, 9000)
+    # The intent is to get a free port between 8000 and 9000, like for the HTTP
+    # server. But we already know that port 8000..PORT are not available. So
+    # starting from `PORT+1` is more efficient.
+    # More importantly though, Windows allows multiple sockets to bind to the
+    # same port in some circumstances (see SO_EXCLUSIVEADDRUSE documentation),
+    # and thus allows the signal socket to bind to the same port as HTTP.
+    # A browser may then end up reaching the wrong socket and causing
+    # fish_config to shutdown prematurely.
+    (sig, sig_port) = create_socket(PORT + 1, 9000)
     threading.Thread(target=capture_enter, args=(sig_port,)).start()
     return sig
 
