@@ -17,34 +17,33 @@ along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
 
-// Delete this once we require Rust 1.74.
-#![allow(unstable_name_collisions)]
-#![allow(clippy::uninlined_format_args)]
-
-#[cfg(feature = "installable")]
-use fish::common::{get_executable_path, wcs2osstring};
-#[allow(unused_imports)]
-use fish::future::IsSomeAnd;
+use cfg_if::cfg_if;
 use fish::{
-    ast::Ast,
-    builtins::shared::{
-        BUILTIN_ERR_MISSING, BUILTIN_ERR_UNKNOWN, STATUS_CMD_OK, STATUS_CMD_UNKNOWN,
+    ast,
+    builtins::{
+        fish_indent, fish_key_reader,
+        shared::{
+            BUILTIN_ERR_MISSING, BUILTIN_ERR_UNEXP_ARG, BUILTIN_ERR_UNKNOWN, STATUS_CMD_ERROR,
+            STATUS_CMD_OK, STATUS_CMD_UNKNOWN,
+        },
     },
     common::{
-        escape, save_term_foreground_process_group, scoped_push_replacer, str2wcstring, wcs2string,
-        PACKAGE_NAME, PROFILING_ACTIVE, PROGRAM_NAME,
+        PACKAGE_NAME, PROFILING_ACTIVE, PROGRAM_NAME, bytes2wcstring, escape,
+        save_term_foreground_process_group, wcs2bytes,
     },
     env::{
-        environment::{env_init, EnvStack, Environment},
-        ConfigPaths, EnvMode, Statuses, CONFIG_PATHS,
+        EnvMode, Statuses,
+        config_paths::ConfigPaths,
+        environment::{EnvStack, Environment, env_init},
     },
     eprintf,
     event::{self, Event},
-    flog::{self, activate_flog_categories_by_pattern, set_flog_file_fd, FLOG, FLOGF},
+    flog::{self, FLOG, FLOGF, activate_flog_categories_by_pattern, set_flog_file_fd},
     fprintf, function, future_feature_flags as features,
     history::{self, start_private_mode},
     io::IoChain,
-    nix::{getpid, isatty},
+    locale::set_libc_locales,
+    nix::{RUsage, getpid, getrusage, isatty},
     panic::panic_handler,
     parse_constants::{ParseErrorList, ParseTreeFlags},
     parse_tree::ParsedSource,
@@ -53,8 +52,8 @@ use fish::{
     path::path_get_config,
     printf,
     proc::{
-        get_login, is_interactive_session, mark_login, mark_no_exec, proc_init,
-        set_interactive_session, Pid,
+        Pid, get_login, is_interactive_session, mark_login, mark_no_exec, proc_init,
+        set_interactive_session,
     },
     reader::{reader_init, reader_read, term_copy_modes},
     signal::{signal_clear_cancel, signal_unblock_all},
@@ -63,137 +62,15 @@ use fish::{
     wchar::prelude::*,
     wutil::waccess,
 };
-use std::ffi::{CString, OsStr, OsString};
+use libc::STDIN_FILENO;
+use std::ffi::{OsStr, OsString};
 use std::fs::File;
-use std::mem::MaybeUninit;
 use std::os::unix::prelude::*;
-use std::path::PathBuf;
-use std::rc::Rc;
-use std::sync::atomic::Ordering;
+use std::path::Path;
 use std::sync::Arc;
+use std::sync::atomic::Ordering;
 use std::{env, ops::ControlFlow};
 
-#[cfg(feature = "installable")]
-// Disable for clippy because otherwise it would require sphinx
-#[cfg(not(clippy))]
-fn install(confirm: bool, dir: &PathBuf) -> bool {
-    use rust_embed::RustEmbed;
-
-    #[derive(RustEmbed)]
-    #[folder = "share/"]
-    struct Asset;
-
-    #[derive(RustEmbed)]
-    #[folder = "target/man/man1"]
-    #[prefix = "man/man1/"]
-    struct Docs;
-
-    use std::fs;
-    use std::io::ErrorKind;
-    use std::io::Write;
-    use std::io::{stderr, stdin};
-
-    // TODO: Translation,
-    // FLOG?
-    // - Install: Translations
-    // - Install: Manpages (build via build.rs)
-    // - Don't install: __fish_build_paths.fish.in
-    if confirm {
-        if isatty(libc::STDIN_FILENO) {
-            eprintln!(
-                "This will write fish's data files to '{}'.\n\
-                 Please enter 'yes' to continue.",
-                dir.display()
-            );
-            eprint!("> ");
-            let _ = stderr().flush();
-        }
-
-        let mut input = String::new();
-        if let Err(error) = stdin().read_line(&mut input) {
-            eprintln!("error: {error}")
-        }
-
-        if input != "yes\n" {
-            eprintln!("Exiting without writing any files\n");
-            return false;
-        }
-    } else {
-        eprintln!("Installing fish's data files to '{}'.", dir.display());
-    }
-
-    // Remove the install directory first, to clean out any removed files.
-    if let Err(err) = fs::remove_dir_all(dir.clone()) {
-        if err.kind() != ErrorKind::NotFound {
-            eprintln!("Removing '{}' failed: {}", dir.display(), err);
-            return false;
-        }
-    }
-
-    // TODO: These are duplicated, no idea how to extract
-    //       them into a function
-    for file in Asset::iter() {
-        let path = dir.join(file.as_ref());
-        let Ok(_) = fs::create_dir_all(path.parent().unwrap()) else {
-            eprintln!(
-                "Creating directory '{}' failed",
-                path.parent().unwrap().display()
-            );
-            return false;
-        };
-        let res = File::create(&path);
-        let Ok(mut f) = res else {
-            eprintln!("Creating file '{}' failed", path.display());
-            continue;
-        };
-        // This should be impossible.
-        let d = Asset::get(&file).expect("File was somehow not included???");
-        if let Err(error) = f.write_all(&d.data) {
-            eprintln!("error: {error}");
-            return false;
-        }
-    }
-
-    for file in Docs::iter() {
-        let path = dir.join(file.as_ref());
-        let Ok(_) = fs::create_dir_all(path.parent().unwrap()) else {
-            eprintln!(
-                "Creating directory '{}' failed",
-                path.parent().unwrap().display()
-            );
-            return false;
-        };
-        let res = File::create(&path);
-        let Ok(mut f) = res else {
-            eprintln!("Creating file '{}' failed", path.display());
-            continue;
-        };
-        // This should be impossible.
-        let d = Docs::get(&file).expect("File was somehow not included???");
-        if let Err(error) = f.write_all(&d.data) {
-            eprintln!("error: {error}");
-            return false;
-        }
-    }
-
-    let verfile = dir.join("fish-install-version");
-    let res = File::create(&verfile);
-    if let Ok(mut f) = res {
-        f.write_all(fish::BUILD_VERSION.as_bytes())
-            .expect("FAILED TO WRITE");
-    } else {
-        eprintln!("Creating file '{}' failed", verfile.display());
-    };
-    return true;
-}
-
-#[cfg(any(clippy, not(feature = "installable")))]
-#[allow(dead_code)]
-fn install(_confirm: bool, _dir: &PathBuf) -> bool {
-    eprintln!("Fish was built without support for self-installation");
-    return false;
-}
-
 /// container to hold the options specified within the command line
 #[derive(Default, Debug)]
 struct FishCmdOpts {
@@ -233,14 +110,8 @@ fn tv_to_msec(tv: &libc::timeval) -> i64 {
 }
 
 fn print_rusage_self() {
-    let mut rs = MaybeUninit::uninit();
-    if unsafe { libc::getrusage(libc::RUSAGE_SELF, rs.as_mut_ptr()) } != 0 {
-        let s = CString::new("getrusage").unwrap();
-        unsafe { libc::perror(s.as_ptr()) }
-        return;
-    }
-    let rs: libc::rusage = unsafe { rs.assume_init() };
-    let rss_kb = if cfg!(target_os = "macos") {
+    let rs = getrusage(RUsage::RSelf);
+    let rss_kb = if cfg!(apple) {
         // mac use bytes.
         rs.ru_maxrss / 1024
     } else {
@@ -253,12 +124,12 @@ fn print_rusage_self() {
     let total_time = user_time + sys_time;
     let signals = rs.ru_nsignals;
 
-    eprintln!("  rusage self:");
-    eprintln!("      user time: {sys_time} ms");
-    eprintln!("       sys time: {user_time} ms");
-    eprintln!("     total time: {total_time} ms");
-    eprintln!("        max rss: {rss_kb} kb");
-    eprintln!("        signals: {signals}");
+    eprintf!("  rusage self:\n");
+    eprintf!("      user time: %s ms\n", sys_time.to_string());
+    eprintf!("       sys time: %s ms\n", user_time.to_string());
+    eprintf!("     total time: %s ms\n", total_time.to_string());
+    eprintf!("        max rss: %s kb\n", rss_kb.to_string());
+    eprintf!("        signals: %s\n", signals.to_string());
 }
 
 // Source the file config.fish in the given directory.
@@ -276,7 +147,7 @@ fn source_config_in_directory(parser: &Parser, dir: &wstr) -> bool {
     if waccess(&config_pathname, libc::R_OK) != 0 {
         FLOGF!(
             config,
-            "not sourcing %ls (not readable or does not exist)",
+            "not sourcing %s (not readable or does not exist)",
             escaped_pathname
         );
         return false;
@@ -288,82 +159,46 @@ fn source_config_in_directory(parser: &Parser, dir: &wstr) -> bool {
     parser.libdata_mut().within_fish_init = true;
     let _ = parser.eval(&cmd, &IoChain::new());
     parser.libdata_mut().within_fish_init = false;
-    return true;
-}
-
-#[cfg(feature = "installable")]
-fn check_version_file(paths: &ConfigPaths, datapath: &wstr) -> Option<bool> {
-    // (false-positive, is_none_or is a backport, this builds with 1.70)
-    #[allow(clippy::incompatible_msrv)]
-    if paths
-        .bin
-        .clone()
-        .is_none_or(|x| !x.starts_with(env!("CARGO_MANIFEST_DIR")))
-    {
-        // When fish is installable, we write the version to a file,
-        // now we check it.
-        let verfile = PathBuf::from(wcs2osstring(datapath)).join("fish-install-version");
-        let version = std::fs::read_to_string(verfile).ok()?;
-
-        return Some(version == fish::BUILD_VERSION);
-    }
-    // When running from the manifest dir, we'll just run.
-    return Some(true);
+    true
 }
 
 /// Parse init files. exec_path is the path of fish executable as determined by argv[0].
 fn read_init(parser: &Parser, paths: &ConfigPaths) {
-    let datapath = str2wcstring(paths.data.as_os_str().as_bytes());
-
-    #[cfg(feature = "installable")]
-    {
-        // If the version file is non-existent or out of date,
-        // we try to install automatically, but only if we're interactive.
-        // If we're not interactive, we still print an error later on pointing to `--install` if they don't exist,
-        // but don't complain if they're merely out-of-date.
-        // We do specifically check for a tty because we want to read input to confirm.
-        let v = check_version_file(paths, &datapath);
-
-        #[allow(clippy::incompatible_msrv)]
-        if v.is_none_or(|x| !x) && is_interactive_session() && isatty(libc::STDIN_FILENO) {
-            if v.is_none() {
-                FLOG!(
-                    warning,
-                    "Fish's asset files are missing. Trying to install them."
+    cfg_if!(
+        if #[cfg(feature = "embed-data")] {
+            use fish::autoload::Asset;
+            let emfile = Asset::get("config.fish").expect("Embedded file not found");
+            let src = bytes2wcstring(&emfile.data);
+            parser.libdata_mut().within_fish_init = true;
+            let fname: Arc<WString> = Arc::new(L!("embedded:config.fish").into());
+            let ret = parser.eval_file_wstr(src, fname, &IoChain::new(), None);
+            parser.libdata_mut().within_fish_init = false;
+            if let Err(msg) = ret {
+                eprintf!("%s", msg);
+            }
+        } else {
+            let datapath = bytes2wcstring(paths.data.as_ref().unwrap().as_os_str().as_bytes());
+            if !source_config_in_directory(parser, &datapath) {
+                // If we cannot read share/config.fish, our internal configuration,
+                // something is wrong.
+                // That also means that our functions won't be found,
+                // and so any config we get would almost certainly be broken.
+                let escaped_pathname = escape(&datapath);
+                FLOGF!(
+                    error,
+                    "Fish cannot find its asset files in '%s'.\n\
+                     Refusing to read configuration because of this.",
+                    escaped_pathname,
                 );
-            } else {
-                FLOG!(
-                    warning,
-                    "Fish's asset files are out of date. Trying to install them."
-                );
+                return;
             }
-
-            install(true, &PathBuf::from(wcs2osstring(&datapath)));
-            // We try to go on if installation failed (or was rejected) here
-            // If the assets are missing, we will trigger a later error,
-            // if they are outdated, things will probably (tm) work somewhat.
         }
-    }
-    if !source_config_in_directory(parser, &datapath) {
-        // If we cannot read share/config.fish, our internal configuration,
-        // something is wrong.
-        // That also means that our functions won't be found,
-        // and so any config we get would almost certainly be broken.
-        let escaped_pathname = escape(&datapath);
-        FLOGF!(
-            error,
-            "Fish cannot find its asset files in '%ls'.\n\
-             Refusing to read configuration because of this.",
-            escaped_pathname,
-        );
-        #[cfg(feature = "installable")]
-        FLOG!(
-            error,
-            "If you installed via `cargo install`, please run `fish --install` and restart fish."
-        );
-        return;
-    }
-    source_config_in_directory(parser, &str2wcstring(paths.sysconf.as_os_str().as_bytes()));
+    );
+
+    source_config_in_directory(
+        parser,
+        &bytes2wcstring(paths.sysconf.as_os_str().as_bytes()),
+    );
 
     // We need to get the configuration directory before we can source the user configuration file.
     // If path_get_config returns false then we have no configuration directory and no custom config
@@ -373,13 +208,13 @@ fn read_init(parser: &Parser, paths: &ConfigPaths) {
     }
 }
 
-fn run_command_list(parser: &Parser, cmds: &[OsString]) -> i32 {
-    let mut retval = STATUS_CMD_OK;
+fn run_command_list(parser: &Parser, cmds: &[OsString]) -> Result<(), libc::c_int> {
+    let mut retval = Ok(());
     for cmd in cmds {
-        let cmd_wcs = str2wcstring(cmd.as_bytes());
+        let cmd_wcs = bytes2wcstring(cmd.as_bytes());
 
         let mut errors = ParseErrorList::new();
-        let ast = Ast::parse(&cmd_wcs, ParseTreeFlags::empty(), Some(&mut errors));
+        let ast = ast::parse(&cmd_wcs, ParseTreeFlags::empty(), Some(&mut errors));
         let errored = ast.errored() || {
             parse_util_detect_errors_in_ast(&ast, &cmd_wcs, Some(&mut errors)).is_err()
         };
@@ -387,27 +222,27 @@ fn run_command_list(parser: &Parser, cmds: &[OsString]) -> i32 {
         if !errored {
             // Construct a parsed source ref.
             let ps = Arc::new(ParsedSource::new(cmd_wcs, ast));
-            let _ = parser.eval_parsed_source(&ps, &IoChain::new(), None, BlockType::top);
-            retval = STATUS_CMD_OK;
+            let _ = parser.eval_parsed_source(&ps, &IoChain::new(), None, BlockType::top, false);
+            retval = Ok(());
         } else {
             let backtrace = parser.get_backtrace(&cmd_wcs, &errors);
             eprintf!("%s", backtrace);
             // XXX: Why is this the return for "unknown command"?
-            retval = STATUS_CMD_UNKNOWN;
+            retval = Err(STATUS_CMD_UNKNOWN);
         }
     }
 
-    retval.unwrap()
+    retval
 }
 
 fn fish_parse_opt(args: &mut [WString], opts: &mut FishCmdOpts) -> ControlFlow<i32, usize> {
-    use fish::wgetopt::{wopt, ArgType::*, WGetopter, WOption};
+    use fish::wgetopt::{ArgType::*, WGetopter, WOption, wopt};
 
     const RUSAGE_ARG: char = 1 as char;
     const PRINT_DEBUG_CATEGORIES_ARG: char = 2 as char;
     const PROFILE_STARTUP_ARG: char = 3 as char;
 
-    const SHORT_OPTS: &wstr = L!("+:hPilNnvc:C:p:d:f:D:o:");
+    const SHORT_OPTS: &wstr = L!("+hPilNnvc:C:p:d:f:D:o:");
     const LONG_OPTS: &[WOption<'static>] = &[
         wopt(L!("command"), RequiredArgument, 'c'),
         wopt(L!("init-command"), RequiredArgument, 'C'),
@@ -420,7 +255,6 @@ fn fish_parse_opt(args: &mut [WString], opts: &mut FishCmdOpts) -> ControlFlow<i
         wopt(L!("no-config"), NoArgument, 'N'),
         wopt(L!("no-execute"), NoArgument, 'n'),
         wopt(L!("print-rusage-self"), NoArgument, RUSAGE_ARG),
-        wopt(L!("install"), OptionalArgument, 'I'),
         wopt(
             L!("print-debug-categories"),
             NoArgument,
@@ -439,77 +273,22 @@ fn fish_parse_opt(args: &mut [WString], opts: &mut FishCmdOpts) -> ControlFlow<i
         match c {
             'c' => opts
                 .batch_cmds
-                .push(OsString::from_vec(wcs2string(w.woptarg.unwrap()))),
+                .push(OsString::from_vec(wcs2bytes(w.woptarg.unwrap()))),
             'C' => opts
                 .postconfig_cmds
-                .push(OsString::from_vec(wcs2string(w.woptarg.unwrap()))),
+                .push(OsString::from_vec(wcs2bytes(w.woptarg.unwrap()))),
             'd' => {
                 activate_flog_categories_by_pattern(w.woptarg.unwrap());
                 for cat in flog::categories::all_categories() {
                     if cat.enabled.load(Ordering::Relaxed) {
-                        println!("Debug enabled for category: {}", cat.name);
+                        printf!("Debug enabled for category: %s\n", cat.name);
                     }
                 }
             }
-            'o' => opts.debug_output = Some(OsString::from_vec(wcs2string(w.woptarg.unwrap()))),
+            'o' => opts.debug_output = Some(OsString::from_vec(wcs2bytes(w.woptarg.unwrap()))),
             'f' => opts.features = w.woptarg.unwrap().to_owned(),
             'h' => opts.batch_cmds.push("__fish_print_help fish".into()),
             'i' => opts.is_interactive_session = true,
-            'I' => {
-                #[cfg(not(feature = "installable"))]
-                eprintln!("Fish was built without support for self-installation");
-                #[cfg(feature = "installable")]
-                if let Some(path) = w.woptarg {
-                    // We were given an explicit path.
-                    // Install us there as a relocatable install.
-                    // That means:
-                    // path/bin/fish is the fish binary
-                    // path/share/fish/ is the data directory
-                    // path/etc/fish is sysconf????
-                    use std::fs;
-                    use std::path::Path;
-                    let dir = PathBuf::from(wcs2osstring(path));
-                    if install(true, &dir.join("share/fish/install")) {
-                        for sub in &["share/fish/install", "etc/fish", "bin"] {
-                            let p = dir.join(sub);
-                            let Ok(_) = fs::create_dir_all(p.clone()) else {
-                                eprintln!("Creating directory '{}' failed", p.display());
-                                std::process::exit(1);
-                            };
-                        }
-
-                        // Copy ourselves there.
-                        let argv0 = OsString::from_vec(wcs2string(&args[0]));
-                        let exec_path =
-                            get_executable_path(<OsString as AsRef<Path>>::as_ref(&argv0));
-                        let binpath = dir.join("bin/fish");
-                        if let Ok(exec_path) = exec_path.canonicalize() {
-                            if exec_path != binpath {
-                                if let Err(err) = std::fs::copy(exec_path, binpath.clone()) {
-                                    FLOG!(error, "Cannot copy fish to", binpath.display());
-                                    FLOG!(error, err);
-                                    std::process::exit(1);
-                                }
-                                println!(
-                                    "Fish installed in '{}'. Start that from now on.",
-                                    binpath.display()
-                                );
-                                // TODO: Reexec fish?
-                                std::process::exit(0);
-                            }
-                        } else {
-                            FLOG!(error, "Cannot copy fish to '%ls'. Please copy the fish binary there manually", binpath.display());
-                        }
-                    }
-                } else {
-                    let paths = Some(&*CONFIG_PATHS);
-                    let Some(paths) = paths else {
-                        FLOG!(error, "Cannot find config paths");
-                        std::process::exit(1);
-                    };
-                    install(true, &paths.data);
-                }
-            }
             'l' => opts.is_login = true,
             'N' => {
                 opts.no_config = true;
@@ -528,20 +307,19 @@ fn fish_parse_opt(args: &mut [WString], opts: &mut FishCmdOpts) -> ControlFlow<i
                 // A little extra space.
                 name_width += 2;
                 for cat in cats.iter() {
-                    let desc = wgettext_str(cat.description);
-                    // https://doc.rust-lang.org/std/fmt/#syntax
+                    let desc = cat.description.localize();
                     // this is left-justified
-                    println!("{:<width$} {}", cat.name, desc, width = name_width);
+                    printf!("%-*s %s\n", name_width, cat.name, desc);
                 }
                 return ControlFlow::Break(0);
             }
             // "--profile" - this does not activate profiling right away,
             // rather it's done after startup is finished.
-            'p' => opts.profile_output = Some(OsString::from_vec(wcs2string(w.woptarg.unwrap()))),
+            'p' => opts.profile_output = Some(OsString::from_vec(wcs2bytes(w.woptarg.unwrap()))),
             PROFILE_STARTUP_ARG => {
                 // With "--profile-startup" we immediately turn profiling on.
                 opts.profile_startup_output =
-                    Some(OsString::from_vec(wcs2string(w.woptarg.unwrap())));
+                    Some(OsString::from_vec(wcs2bytes(w.woptarg.unwrap())));
                 PROFILING_ACTIVE.store(true);
             }
             'P' => opts.enable_private_mode = true,
@@ -557,19 +335,26 @@ fn fish_parse_opt(args: &mut [WString], opts: &mut FishCmdOpts) -> ControlFlow<i
                 // Either remove it or make it work with FLOG.
             }
             '?' => {
-                eprintln!(
-                    "{}",
+                eprintf!(
+                    "%s\n",
                     wgettext_fmt!(BUILTIN_ERR_UNKNOWN, "fish", args[w.wopt_index - 1])
                 );
                 return ControlFlow::Break(1);
             }
             ':' => {
-                eprintln!(
-                    "{}",
+                eprintf!(
+                    "%s\n",
                     wgettext_fmt!(BUILTIN_ERR_MISSING, "fish", args[w.wopt_index - 1])
                 );
                 return ControlFlow::Break(1);
             }
+            ';' => {
+                eprintf!(
+                    "%s\n",
+                    wgettext_fmt!(BUILTIN_ERR_UNEXP_ARG, "fish", args[w.wopt_index - 1])
+                );
+                return ControlFlow::Break(1);
+            }
             _ => panic!("unexpected retval from WGetopter"),
         }
     }
@@ -581,7 +366,7 @@ fn fish_parse_opt(args: &mut [WString], opts: &mut FishCmdOpts) -> ControlFlow<i
     // We are an interactive session if we have not been given an explicit
     // command or file to execute and stdin is a tty. Note that the -i or
     // --interactive options also force interactive mode.
-    if opts.batch_cmds.is_empty() && optind == args.len() && isatty(libc::STDIN_FILENO) {
+    if opts.batch_cmds.is_empty() && optind == args.len() && isatty(STDIN_FILENO) {
         set_interactive_session(true);
     }
 
@@ -589,6 +374,15 @@ fn fish_parse_opt(args: &mut [WString], opts: &mut FishCmdOpts) -> ControlFlow<i
 }
 
 fn main() {
+    // If we are called as "/path/to/fish_key_reader", become fish_key_reader.
+    if let Some(name) = env::args_os().next() {
+        let p = Path::new(&name).file_name().and_then(|x| x.to_str());
+        if p == Some("fish_key_reader") {
+            return fish_key_reader::main();
+        } else if p == Some("fish_indent") {
+            return fish_indent::main();
+        }
+    }
     PROGRAM_NAME.set(L!("fish")).unwrap();
     if !cfg!(small_main_stack) {
         panic_handler(throwing_main);
@@ -601,33 +395,29 @@ fn main() {
 }
 
 fn throwing_main() -> i32 {
-    let mut res = 1;
+    let mut res = Err(STATUS_CMD_ERROR);
 
     signal_unblock_all();
     topic_monitor::topic_monitor_init();
     threads::init();
 
-    {
-        let s = CString::new("").unwrap();
-        unsafe {
-            libc::setlocale(libc::LC_ALL, s.as_ptr());
-        }
-    }
+    // Safety: single-threaded.
+    unsafe {
+        set_libc_locales(/*log_ok=*/ false)
+    };
 
-    let mut args: Vec<WString> = env::args_os()
-        .map(|osstr| str2wcstring(osstr.as_bytes()))
-        .collect();
-    if args.is_empty() {
-        args.push("fish".into());
-    }
+    fish::wutil::gettext::initialize_gettext();
 
     // Enable debug categories set in FISH_DEBUG.
     // This is in *addition* to the ones given via --debug.
     if let Some(debug_categories) = env::var_os("FISH_DEBUG") {
-        let s = str2wcstring(debug_categories.as_bytes());
+        let s = bytes2wcstring(debug_categories.as_bytes());
         activate_flog_categories_by_pattern(&s);
     }
 
+    let mut args: Vec<WString> = env::args_os()
+        .map(|osstr| bytes2wcstring(osstr.as_bytes()))
+        .collect();
     let mut opts = FishCmdOpts::default();
     let mut my_optind = match fish_parse_opt(&mut args, &mut opts) {
         ControlFlow::Continue(optind) => optind,
@@ -646,7 +436,7 @@ fn throwing_main() -> i32 {
             .write(true)
             .truncate(true)
             .create(true)
-            .open(debug_path.clone())
+            .open(&debug_path)
         {
             Ok(dbg_file) => {
                 // Rust sets O_CLOEXEC by default
@@ -654,9 +444,10 @@ fn throwing_main() -> i32 {
                 set_flog_file_fd(dbg_file.into_raw_fd());
             }
             Err(e) => {
-                // TODO: should not be debug-print
-                eprintln!("Could not open file {:?}", debug_path);
-                eprintln!("{}", e);
+                let debug_path_string = format!("{debug_path:?}");
+                // TODO: should be translated
+                eprintf!("Could not open file %s\n", debug_path_string);
+                eprintf!("%s\n", e);
                 return 1;
             }
         };
@@ -691,16 +482,15 @@ fn throwing_main() -> i32 {
         save_term_foreground_process_group();
     }
 
-    let mut paths: Option<&ConfigPaths> = None;
     // If we're not executing, there's no need to find the config.
-    if !opts.no_exec {
-        paths = Some(&*CONFIG_PATHS);
+    let config_paths = if !opts.no_exec {
+        let config_paths = ConfigPaths::new();
         env_init(
-            paths,
+            Some(&config_paths),
             /* do uvars */ !opts.no_config,
             /* default paths */ opts.no_config,
         );
-        paths
+        Some(config_paths)
     } else {
         None
     };
@@ -714,17 +504,17 @@ fn throwing_main() -> i32 {
         }
     }
     features::set_from_string(opts.features.as_utfstr());
+    fish::env_dispatch::read_terminfo_database(EnvStack::globals());
     proc_init();
-    fish::env::misc_init();
     reader_init(true);
 
     // Construct the root parser!
-    let env = Rc::new(EnvStack::globals().create_child(true /* dispatches_var_changes */));
+    let env = EnvStack::globals().create_child(true /* dispatches_var_changes */);
     let parser = &Parser::new(env, CancelBehavior::Clear);
     parser.set_syncs_uvars(!opts.no_config);
 
     if !opts.no_exec && !opts.no_config {
-        read_init(parser, paths.as_ref().unwrap());
+        read_init(parser, config_paths.as_ref().unwrap());
     }
 
     if is_interactive_session() && opts.no_config && !opts.no_exec {
@@ -735,7 +525,7 @@ fn throwing_main() -> i32 {
             L!("fish_default_key_bindings").to_owned(),
         );
         if function::exists(L!("fish_default_key_bindings"), parser) {
-            run_command_list(parser, &[OsString::from("fish_default_key_bindings")]);
+            let _ = run_command_list(parser, &[OsString::from("fish_default_key_bindings")]);
         }
     }
 
@@ -743,7 +533,7 @@ fn throwing_main() -> i32 {
     term_copy_modes();
 
     // Stomp the exit status of any initialization commands (issue #635).
-    parser.set_last_statuses(Statuses::just(STATUS_CMD_OK.unwrap()));
+    parser.set_last_statuses(Statuses::just(STATUS_CMD_OK));
 
     // TODO: if-let-chains
     if opts.profile_startup_output.is_some() && opts.profile_startup_output != opts.profile_output {
@@ -793,7 +583,7 @@ fn throwing_main() -> i32 {
         }
         res = reader_read(parser, libc::STDIN_FILENO, &IoChain::new());
     } else {
-        let n = wcs2string(&args[my_optind]);
+        let n = wcs2bytes(&args[my_optind]);
         let path = OsStr::from_bytes(&n);
         my_optind += 1;
         // Rust sets cloexec by default, see above
@@ -805,7 +595,7 @@ fn throwing_main() -> i32 {
                     wgettext!("Error reading script file '%s':"),
                     path.to_string_lossy()
                 );
-                eprintln!("{}", e);
+                eprintf!("%s\n", e);
             }
             Ok(f) => {
                 let list = &args[my_optind..];
@@ -815,17 +605,16 @@ fn throwing_main() -> i32 {
                     list.iter().map(|s| s.to_owned()).collect(),
                 );
                 let rel_filename = &args[my_optind - 1];
-                let _filename_push = scoped_push_replacer(
-                    |new_value| {
-                        std::mem::replace(&mut parser.libdata_mut().current_filename, new_value)
-                    },
-                    Some(Arc::new(rel_filename.to_owned())),
-                );
+                let _filename_push = parser
+                    .library_data
+                    .scoped_set(Some(Arc::new(rel_filename.to_owned())), |s| {
+                        &mut s.current_filename
+                    });
                 res = reader_read(parser, f.as_raw_fd(), &IoChain::new());
-                if res != 0 {
+                if res.is_err() {
                     FLOGF!(
                         warning,
-                        wgettext!("Error while reading file %ls\n"),
+                        wgettext!("Error while reading file %s\n"),
                         path.to_string_lossy()
                     );
                 }
@@ -833,16 +622,13 @@ fn throwing_main() -> i32 {
         }
     }
 
-    let exit_status = if res != 0 {
-        STATUS_CMD_UNKNOWN.unwrap()
+    let exit_status = if res.is_err() {
+        STATUS_CMD_UNKNOWN
     } else {
         parser.get_last_status()
     };
 
-    event::fire(
-        parser,
-        Event::process_exit(Pid::new(getpid()).unwrap(), exit_status),
-    );
+    event::fire(parser, Event::process_exit(Pid::new(getpid()), exit_status));
 
     // Trigger any exit handlers.
     event::fire_generic(
