@@ -2,12 +2,12 @@
 
 use super::prelude::*;
 use crate::common::{
-    escape, escape_string, str2wcstring, valid_var_name, EscapeFlags, EscapeStringStyle,
+    EscapeFlags, EscapeStringStyle, bytes2wcstring, escape, escape_string, valid_var_name,
 };
 use crate::highlight::{colorize, highlight_shell};
-use crate::input::{input_function_get_names, input_mappings, InputMappingSet, KeyNameStyle};
+use crate::input::{InputMappingSet, KeyNameStyle, input_function_get_names, input_mappings};
 use crate::key::{
-    self, char_to_symbol, function_key, parse_keys, Key, Modifiers, KEY_NAMES, MAX_FUNCTION_KEY,
+    self, KEY_NAMES, Key, MAX_FUNCTION_KEY, Modifiers, char_to_symbol, function_key, parse_keys,
 };
 use crate::nix::isatty;
 use std::sync::MutexGuard;
@@ -157,7 +157,7 @@ impl BuiltinBind {
             let mut colors = Vec::new();
             highlight_shell(&out, &mut colors, &parser.context(), false, None);
             let colored = colorize(&out, &colors, parser.vars());
-            streams.out.append(str2wcstring(&colored));
+            streams.out.append(bytes2wcstring(&colored));
         } else {
             streams.out.append(out);
         }
@@ -350,13 +350,13 @@ impl BuiltinBind {
                 if !self.opts.silent {
                     if seq.len() == 1 {
                         streams.err.append(wgettext_fmt!(
-                            "%ls: No binding found for key '%ls'\n",
+                            "%s: No binding found for key '%s'\n",
                             cmd,
                             seq[0]
                         ));
                     } else {
                         streams.err.append(wgettext_fmt!(
-                            "%ls: No binding found for key sequence '%ls'\n",
+                            "%s: No binding found for key sequence '%s'\n",
                             cmd,
                             eseq
                         ));
@@ -432,7 +432,7 @@ fn parse_cmd_opts(
             'h' => opts.print_help = true,
             'k' => {
                 streams.err.append(wgettext_fmt!(
-                    "%ls: the -k/--key syntax is no longer supported. See `bind --help` and `bind --key-names`\n",
+                    "%s: the -k/--key syntax is no longer supported. See `bind --help` and `bind --key-names`\n",
                     cmd,
                 ));
                 return Err(STATUS_INVALID_ARGS);
@@ -524,25 +524,25 @@ impl BuiltinBind {
         match self.opts.mode {
             BIND_ERASE => {
                 // If we get both, we erase both.
-                if self.opts.user {
-                    if self.erase(
+                if self.opts.user
+                    && self.erase(
                         &argv[optind..],
                         self.opts.all,
                         true, /* user */
                         streams,
-                    ) {
-                        return Err(STATUS_CMD_ERROR);
-                    }
+                    )
+                {
+                    return Err(STATUS_CMD_ERROR);
                 }
-                if self.opts.preset {
-                    if self.erase(
+                if self.opts.preset
+                    && self.erase(
                         &argv[optind..],
                         self.opts.all,
                         false, /* user */
                         streams,
-                    ) {
-                        return Err(STATUS_CMD_ERROR);
-                    }
+                    )
+                {
+                    return Err(STATUS_CMD_ERROR);
                 }
             }
             BIND_INSERT => {
@@ -555,7 +555,7 @@ impl BuiltinBind {
             _ => {
                 streams
                     .err
-                    .append(wgettext_fmt!("%ls: Invalid state\n", cmd));
+                    .append(wgettext_fmt!("%s: Invalid state\n", cmd));
                 return Err(STATUS_CMD_ERROR);
             }
         }
