@@ -1,6 +1,5 @@
 use super::prelude::*;
-use crate::common;
-use crate::future_feature_flags::{feature_test, FeatureFlag};
+use crate::future_feature_flags::{FeatureFlag, feature_test};
 use crate::should_flog;
 
 mod test_expressions {
@@ -8,8 +7,8 @@ mod test_expressions {
 
     use crate::nix::isatty;
     use crate::wutil::{
-        file_id_for_path, fish_wcswidth, lwstat, waccess, wcstod::wcstod, wcstoi_opts, wstat,
-        Error, Options,
+        Error, Options, file_id_for_path, fish_wcswidth, lwstat, waccess, wcstod::wcstod,
+        wcstoi_opts, wstat,
     };
     use once_cell::sync::Lazy;
     use std::collections::HashMap;
@@ -86,7 +85,7 @@ mod test_expressions {
             d // "-d", for directories
             e // "-e", for files that exist
             f // "-f", for for regular files
-            G // "-G", for check effective group id
+            G // "-G", for check effective group ID
             g // "-g", for set-group-id
             k // "-k", for sticky bit
             O // "-O", for check effective user id
@@ -757,14 +756,14 @@ mod test_expressions {
                 if !parser.errors.is_empty() {
                     err.push_utfstr(&parser.errors[0]);
                 } else {
-                    sprintf!(=> err, "unexpected argument at index %lu: '%ls'",
+                    sprintf!(=> err, "unexpected argument at index %u: '%s'",
                              result.as_ref().unwrap().range().end + 1,
                              args[result.as_ref().unwrap().range().end]);
                 }
                 err.push('\n');
                 err.push_utfstr(&commandline);
                 err.push('\n');
-                err.push_utfstr(&sprintf!("%*ls%ls\n", len_to_err + 1, " ", "^"));
+                err.push_utfstr(&sprintf!("%*s%s\n", len_to_err + 1, " ", "^"));
             }
 
             if result.is_some() {
@@ -846,12 +845,12 @@ mod test_expressions {
                 if let Ok(prefix_int) = wcstoi_opts(arg, options) {
                     let _: i64 = prefix_int; // to help type inference
                     errors.push(wgettext_fmt!(
-                        "Integer %lld in '%ls' followed by non-digit",
+                        "Integer %d in '%s' followed by non-digit",
                         prefix_int,
                         arg
                     ));
                 } else {
-                    errors.push(wgettext_fmt!("Argument is not a number: '%ls'", arg));
+                    errors.push(wgettext_fmt!("Argument is not a number: '%s'", arg));
                 }
             } else if floating.is_ok_and(|x| x.is_nan()) {
                 // NaN is an error as far as we're concerned.
@@ -859,9 +858,9 @@ mod test_expressions {
             } else if floating.is_ok_and(|x| x.is_infinite()) {
                 errors.push(wgettext!("Number is infinite").to_owned());
             } else if integral == Err(Error::Overflow) {
-                errors.push(wgettext_fmt!("Result too large: %ls", arg));
+                errors.push(wgettext_fmt!("Result too large: %s", arg));
             } else {
-                errors.push(wgettext_fmt!("Invalid number: %ls", arg));
+                errors.push(wgettext_fmt!("Invalid number: %s", arg));
             }
             false
         }
@@ -931,7 +930,7 @@ mod test_expressions {
                     StatPredicate::e => true,
                     // "-f", for regular files
                     StatPredicate::f => md.file_type().is_file(),
-                    // "-G", for check effective group id
+                    // "-G", for check effective group ID
                     StatPredicate::G => md.gid() == crate::nix::getegid(),
                     // "-g", for set-group-id
                     StatPredicate::g => md.permissions().mode() & S_ISGID != 0,
@@ -1023,7 +1022,7 @@ pub fn test(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Bui
     if feature_test(FeatureFlag::test_require_arg) {
         if argc == 0 {
             streams.err.appendln(wgettext_fmt!(
-                "%ls: Expected at least one argument",
+                "%s: Expected at least one argument",
                 program_name
             ));
             builtin_print_error_trailer(parser, streams.err, program_name);
@@ -1038,21 +1037,19 @@ pub fn test(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Bui
     } else if argc == 0 {
         if should_flog!(deprecated_test) {
             streams.err.appendln(wgettext_fmt!(
-                "%ls: called with no arguments. This will be an error in future.",
+                "%s: called with no arguments. This will be an error in future.",
                 program_name
             ));
             streams.err.append(parser.current_line());
         }
         return Err(STATUS_INVALID_ARGS); // Per 1003.1, exit false.
     } else if argc == 1 {
-        if should_flog!(deprecated_test) {
-            if args[0] != "-z" {
-                streams.err.appendln(wgettext_fmt!(
-                    "%ls: called with one argument. This will return false in future.",
-                    program_name
-                ));
-                streams.err.append(parser.current_line());
-            }
+        if should_flog!(deprecated_test) && args[0] != "-z" {
+            streams.err.appendln(wgettext_fmt!(
+                "%s: called with one argument. This will return false in future.",
+                program_name
+            ));
+            streams.err.append(parser.current_line());
         }
         // Per 1003.1, exit true if the arg is non-empty.
         return if args[0].is_empty() {
@@ -1074,14 +1071,12 @@ pub fn test(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Bui
     let mut eval_errors = Vec::new();
     let result = expr.evaluate(streams, &mut eval_errors);
     if !eval_errors.is_empty() {
-        if !common::should_suppress_stderr_for_tests() {
-            for eval_error in eval_errors {
-                streams.err.appendln(&eval_error);
-            }
-            // Add a backtrace but not the "see help" message
-            // because this isn't about passing the wrong options.
-            streams.err.append(parser.current_line());
+        for eval_error in eval_errors {
+            streams.err.appendln(&eval_error);
         }
+        // Add a backtrace but not the "see help" message
+        // because this isn't about passing the wrong options.
+        streams.err.append(parser.current_line());
         return Err(STATUS_INVALID_ARGS);
     }
 
@@ -1091,3 +1086,185 @@ pub fn test(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Bui
         Err(STATUS_CMD_ERROR)
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::test as builtin_test;
+    use crate::builtins::prelude::*;
+    use crate::io::{IoChain, OutputStream};
+    use crate::tests::prelude::*;
+
+    fn run_one_test_test_mbracket(expected: i32, lst: &[&str], bracket: bool) -> bool {
+        let parser = TestParser::new();
+        let mut argv = Vec::new();
+        if bracket {
+            argv.push(L!("[").to_owned());
+        } else {
+            argv.push(L!("test").to_owned());
+        }
+        for s in lst {
+            argv.push(WString::from_str(s));
+        }
+        if bracket {
+            argv.push(L!("]").to_owned())
+        };
+
+        // Convert to &[&wstr].
+        let mut argv = argv.iter().map(|s| s.as_ref()).collect::<Vec<_>>();
+        let mut out = OutputStream::Null;
+        let mut err = OutputStream::Null;
+        let io_chain = IoChain::new();
+        let mut streams = IoStreams::new(&mut out, &mut err, &io_chain);
+
+        let result = builtin_test(&parser, &mut streams, &mut argv).builtin_status_code();
+
+        if result != expected {
+            eprintf!(
+                "expected builtin_test() to return %s, got %s\n",
+                expected.to_string(),
+                result.to_string()
+            );
+        }
+        result == expected
+    }
+
+    fn run_test_test(expected: i32, lst: &[&str]) -> bool {
+        let nobracket = run_one_test_test_mbracket(expected, lst, false);
+        let bracket = run_one_test_test_mbracket(expected, lst, true);
+        assert_eq!(nobracket, bracket);
+        nobracket
+    }
+
+    fn test_test_brackets() {
+        // Ensure [ knows it needs a ].
+        let parser = TestParser::new();
+
+        let mut out = OutputStream::Null;
+        let mut err = OutputStream::Null;
+        let io_chain = IoChain::new();
+        let mut streams = IoStreams::new(&mut out, &mut err, &io_chain);
+
+        let args1 = &mut [L!("["), L!("foo")];
+        assert_eq!(
+            builtin_test(&parser, &mut streams, args1),
+            Err(STATUS_INVALID_ARGS)
+        );
+
+        let args2 = &mut [L!("["), L!("foo"), L!("]")];
+        assert_eq!(builtin_test(&parser, &mut streams, args2), Ok(SUCCESS));
+
+        let args3 = &mut [L!("["), L!("foo"), L!("]"), L!("bar")];
+        assert_eq!(
+            builtin_test(&parser, &mut streams, args3),
+            Err(STATUS_INVALID_ARGS)
+        );
+    }
+
+    #[rustfmt::skip]
+    fn test_test() {
+        assert!(run_test_test(0, &["5", "-ne", "6"]));
+        assert!(run_test_test(0, &["5", "-eq", "5"]));
+        assert!(run_test_test(0, &["0", "-eq", "0"]));
+        assert!(run_test_test(0, &["-1", "-eq", "-1"]));
+        assert!(run_test_test(0, &["1", "-ne", "-1"]));
+        assert!(run_test_test(1, &[" 2 ", "-ne", "2"]));
+        assert!(run_test_test(0, &[" 2", "-eq", "2"]));
+        assert!(run_test_test(0, &["2 ", "-eq", "2"]));
+        assert!(run_test_test(0, &[" 2 ", "-eq", "2"]));
+        assert!(run_test_test(2, &[" 2x", "-eq", "2"]));
+        assert!(run_test_test(2, &["", "-eq", "0"]));
+        assert!(run_test_test(2, &["", "-ne", "0"]));
+        assert!(run_test_test(2, &["  ", "-eq", "0"]));
+        assert!(run_test_test(2, &["  ", "-ne", "0"]));
+        assert!(run_test_test(2, &["x", "-eq", "0"]));
+        assert!(run_test_test(2, &["x", "-ne", "0"]));
+        assert!(run_test_test(1, &["-1", "-ne", "-1"]));
+        assert!(run_test_test(0, &["abc", "!=", "def"]));
+        assert!(run_test_test(1, &["abc", "=", "def"]));
+        assert!(run_test_test(0, &["5", "-le", "10"]));
+        assert!(run_test_test(0, &["10", "-le", "10"]));
+        assert!(run_test_test(1, &["20", "-le", "10"]));
+        assert!(run_test_test(0, &["-1", "-le", "0"]));
+        assert!(run_test_test(1, &["0", "-le", "-1"]));
+        assert!(run_test_test(0, &["15", "-ge", "10"]));
+        assert!(run_test_test(0, &["15", "-ge", "10"]));
+        assert!(run_test_test(1, &["!", "15", "-ge", "10"]));
+        assert!(run_test_test(0, &["!", "!", "15", "-ge", "10"]));
+
+        assert!(run_test_test(0, &[
+            "(", "-d", "/", ")",
+            "-o",
+            "(", "!", "-d", "/", ")",
+        ]));
+
+        assert!(run_test_test(0, &["0", "-ne", "1", "-a", "0", "-eq", "0"]));
+        assert!(run_test_test(0, &["0", "-ne", "1", "-a", "-n", "5"]));
+        assert!(run_test_test(0, &["-n", "5", "-a", "10", "-gt", "5"]));
+        assert!(run_test_test(0, &["-n", "3", "-a", "-n", "5"]));
+
+        // Test precedence:
+        //      '0 == 0 || 0 == 1 && 0 == 2'
+        //  should be evaluated as:
+        //      '0 == 0 || (0 == 1 && 0 == 2)'
+        //  and therefore true. If it were
+        //      '(0 == 0 || 0 == 1) && 0 == 2'
+        //  it would be false.
+        assert!(run_test_test(0, &["0", "=", "0", "-o", "0", "=", "1", "-a", "0", "=", "2"]));
+        assert!(run_test_test(0, &["-n", "5", "-o", "0", "=", "1", "-a", "0", "=", "2"]));
+        assert!(run_test_test(1, &["(", "0", "=", "0", "-o", "0", "=", "1", ")", "-a", "0", "=", "2"]));
+        assert!(run_test_test(0, &["0", "=", "0", "-o", "(", "0", "=", "1", "-a", "0", "=", "2", ")"]));
+
+        // A few lame tests for permissions; these need to be a lot more complete.
+        assert!(run_test_test(0, &["-e", "/bin/ls"]));
+        assert!(run_test_test(1, &["-e", "/bin/ls_not_a_path"]));
+        assert!(run_test_test(0, &["-x", "/bin/ls"]));
+        assert!(run_test_test(1, &["-x", "/bin/ls_not_a_path"]));
+        assert!(run_test_test(0, &["-d", "/bin/"]));
+        assert!(run_test_test(1, &["-d", "/bin/ls"]));
+
+        // This failed at one point.
+        assert!(run_test_test(1, &["-d", "/bin", "-a", "5", "-eq", "3"]));
+        assert!(run_test_test(0, &["-d", "/bin", "-o", "5", "-eq", "3"]));
+        assert!(run_test_test(0,&["-d", "/bin", "-a", "!", "5", "-eq", "3"]));
+
+        // We didn't properly handle multiple "just strings" either.
+        assert!(run_test_test(0, &["foo"]));
+        assert!(run_test_test(0, &["foo", "-a", "bar"]));
+
+        // These should be errors.
+        assert!(run_test_test(1, &["foo", "bar"]));
+        assert!(run_test_test(1, &["foo", "bar", "baz"]));
+
+        // This crashed.
+        assert!(run_test_test(1, &["1", "=", "1", "-a", "=", "1"]));
+
+        // Make sure we can treat -S as a parameter instead of an operator.
+        // https://github.com/fish-shell/fish-shell/issues/601
+        assert!(run_test_test(0, &["-S", "=", "-S"]));
+        assert!(run_test_test(1, &["!", "!", "!", "A"]));
+
+        // Verify that 1. doubles are treated as doubles, and 2. integers that cannot be represented as
+        // doubles are still treated as integers.
+        assert!(run_test_test(0, &["4611686018427387904", "-eq", "4611686018427387904"]));
+        assert!(run_test_test(0, &["4611686018427387904.0", "-eq", "4611686018427387904.0"]));
+        assert!(run_test_test(0, &["4611686018427387904.00000000000000001", "-eq", "4611686018427387904.0"]));
+        assert!(run_test_test(1, &["4611686018427387904", "-eq", "4611686018427387905"]));
+        assert!(run_test_test(0, &["-4611686018427387904", "-ne", "4611686018427387904"]));
+        assert!(run_test_test(0, &["-4611686018427387904", "-le", "4611686018427387904"]));
+        assert!(run_test_test(1, &["-4611686018427387904", "-ge", "4611686018427387904"]));
+        assert!(run_test_test(1, &["4611686018427387904", "-gt", "4611686018427387904"]));
+        assert!(run_test_test(0, &["4611686018427387904", "-ge", "4611686018427387904"]));
+
+        // test out-of-range numbers
+        assert!(run_test_test(2, &["99999999999999999999999999", "-ge", "1"]));
+        assert!(run_test_test(2, &["1", "-eq", "-99999999999999999999999999.9"]));
+    }
+
+    #[test]
+    #[serial]
+    fn test_test_builtin() {
+        let _cleanup = test_init();
+        test_test_brackets();
+        test_test();
+    }
+}
