@@ -7,14 +7,14 @@
 use std::sync::atomic::{AtomicBool, AtomicU32, Ordering};
 use std::sync::{Arc, Mutex};
 
-use crate::common::{escape, ScopeGuard};
+use crate::common::{ScopeGuard, escape};
 use crate::flog::FLOG;
 use crate::io::{IoChain, IoStreams};
 use crate::job_group::MaybeJobId;
 use crate::parser::{Block, Parser};
 use crate::proc::Pid;
-use crate::signal::{signal_check_cancel, signal_handle, Signal};
-use crate::termsize;
+use crate::reader::reader_update_termsize;
+use crate::signal::{Signal, signal_check_cancel, signal_handle};
 use crate::wchar::prelude::*;
 
 pub enum event_type_t {
@@ -523,12 +523,10 @@ fn fire_internal(parser: &Parser, event: &Event) {
 
 /// Fire all delayed events attached to the given parser.
 pub fn fire_delayed(parser: &Parser) {
-    {
-        // Do not invoke new event handlers from within event handlers.
-        if parser.scope().is_event {
-            return;
-        };
-    }
+    // Do not invoke new event handlers from within event handlers.
+    if parser.scope().is_event {
+        return;
+    };
 
     // Do not invoke new event handlers if we are unwinding (#6649).
     if signal_check_cancel() != 0 {
@@ -551,7 +549,7 @@ pub fn fire_delayed(parser: &Parser) {
         // HACK: The only variables we change in response to a *signal* are $COLUMNS and $LINES.
         // Do that now.
         if sig == libc::SIGWINCH {
-            termsize::SHARED_CONTAINER.updating(parser);
+            reader_update_termsize(parser)
         }
         let event = Event {
             desc: EventDescription::Signal { signal: sig },
@@ -632,7 +630,7 @@ pub fn print(streams: &mut IoStreams, type_filter: &wstr) {
             }
 
             last_type = std::mem::discriminant(&evt.desc);
-            streams.out.append(sprintf!("Event %ls\n", evt.desc.name()));
+            streams.out.append(sprintf!("Event %s\n", evt.desc.name()));
         }
 
         match &evt.desc {
@@ -640,18 +638,18 @@ pub fn print(streams: &mut IoStreams, type_filter: &wstr) {
                 let name: WString = signal.name().into();
                 streams
                     .out
-                    .append(sprintf!("%ls %ls\n", name, evt.function_name));
+                    .append(sprintf!("%s %s\n", name, evt.function_name));
             }
             EventDescription::ProcessExit { .. } | EventDescription::JobExit { .. } => {}
             EventDescription::CallerExit { .. } => {
                 streams
                     .out
-                    .append(sprintf!("caller-exit %ls\n", evt.function_name));
+                    .append(sprintf!("caller-exit %s\n", evt.function_name));
             }
             EventDescription::Variable { name: param } | EventDescription::Generic { param } => {
                 streams
                     .out
-                    .append(sprintf!("%ls %ls\n", param, evt.function_name));
+                    .append(sprintf!("%s %s\n", param, evt.function_name));
             }
             EventDescription::Any => unreachable!(),
         }
