@@ -1,23 +1,21 @@
 use crate::common::{
-    fish_reserved_codepoint, is_windows_subsystem_for_linux, read_blocked, shell_modes,
-    str2wcstring, WSL,
+    WSL, bytes2wcstring, fish_reserved_codepoint, is_windows_subsystem_for_linux, read_blocked,
+    shell_modes,
 };
 use crate::env::{EnvStack, Environment};
 use crate::fd_readable_set::{FdReadableSet, Timeout};
-use crate::flog::{FloggableDebug, FloggableDisplay, FLOG};
+use crate::flog::{FLOG, FloggableDebug, FloggableDisplay};
 use crate::key::{
-    self, alt, canonicalize_control_char, canonicalize_keyed_control_char, char_to_symbol,
-    function_key, shift, Key, Modifiers, ViewportPosition,
+    self, Key, Modifiers, ViewportPosition, alt, canonicalize_control_char,
+    canonicalize_keyed_control_char, char_to_symbol, function_key, shift,
 };
 use crate::reader::reader_test_and_clear_interrupted;
-use crate::threads::iothread_port;
 use crate::tty_handoff::{
-    maybe_set_kitty_keyboard_capability, maybe_set_scroll_content_up_capability,
-    SCROLL_CONTENT_UP_TERMINFO_CODE, XTVERSION,
+    SCROLL_CONTENT_UP_TERMINFO_CODE, XTVERSION, maybe_set_kitty_keyboard_capability,
+    maybe_set_scroll_content_up_capability,
 };
 use crate::universal_notifier::default_notifier;
 use crate::wchar::{encode_byte_to_char, prelude::*};
-use crate::wutil::encoding::{mbrtowc, mbstate_t, zero_mbstate};
 use crate::wutil::{fish_is_pua, fish_wcstol};
 use std::cell::{RefCell, RefMut};
 use std::collections::VecDeque;
@@ -272,88 +270,6 @@ pub(crate) fn match_key_event_to_key(event: &KeyEvent, key: &Key) -> Option<KeyM
     None
 }
 
-#[test]
-fn test_match_key_event_to_key() {
-    macro_rules! validate {
-        ($evt:expr, $key:expr, $expected:expr) => {
-            assert_eq!(match_key_event_to_key(&$evt, &$key), $expected);
-        };
-    }
-
-    let none = Modifiers::default();
-    let shift = Modifiers::SHIFT;
-    let ctrl = Modifiers::CTRL;
-    let ctrl_shift = Modifiers {
-        ctrl: true,
-        shift: true,
-        ..Default::default()
-    };
-
-    let exact = KeyMatchQuality::Exact;
-    let modulo_shift = KeyMatchQuality::ModuloShift;
-    let base_layout = KeyMatchQuality::BaseLayout;
-    let base_layout_modulo_shift = KeyMatchQuality::BaseLayoutModuloShift;
-
-    validate!(KeyEvent::new(none, 'a'), Key::new(none, 'a'), Some(exact));
-    validate!(KeyEvent::new(none, 'a'), Key::new(none, 'A'), None);
-    validate!(KeyEvent::new(shift, 'a'), Key::new(shift, 'a'), Some(exact));
-    validate!(KeyEvent::new(shift, 'a'), Key::new(none, 'A'), None);
-    validate!(KeyEvent::new(shift, 'ä'), Key::new(none, 'Ä'), None);
-    // For historical reasons we canonicalize notation for ASCII keys like "shift-a" to "A",
-    // but not "shift-a" events - those should send a shifted key.
-    validate!(
-        KeyEvent::new(none, 'A'),
-        Key::new(shift, 'a'),
-        Some(modulo_shift)
-    );
-    validate!(KeyEvent::new(none, 'A'), Key::new(shift, 'A'), None);
-    validate!(KeyEvent::new(none, 'Ä'), Key::new(none, 'Ä'), Some(exact));
-    validate!(KeyEvent::new(none, 'Ä'), Key::new(shift, 'ä'), None);
-
-    // FYI: for codepoints that are not letters with uppercase/lowercase versions, we use
-    // the shifted key in the canonical notation, because the unshifted one may depend on the
-    // keyboard layout.
-    let ctrl_shift_equals = KeyEvent::new_with(ctrl_shift, '=', Some('+'), None);
-    validate!(ctrl_shift_equals, Key::new(ctrl_shift, '='), Some(exact));
-    validate!(ctrl_shift_equals, Key::new(ctrl, '+'), Some(modulo_shift)); // canonical notation
-    validate!(ctrl_shift_equals, Key::new(ctrl_shift, '+'), None);
-    validate!(ctrl_shift_equals, Key::new(ctrl, '='), None);
-
-    // A event like capslock-shift-ä may or may not include a shifted codepoint.
-    //
-    // Without a shifted codepoint, we cannot easily match ctrl-Ä.
-    let caps_ctrl_shift_ä = KeyEvent::new(ctrl_shift, 'ä');
-    validate!(caps_ctrl_shift_ä, Key::new(ctrl_shift, 'ä'), Some(exact)); // canonical notation
-    validate!(caps_ctrl_shift_ä, Key::new(ctrl, 'ä'), None);
-    validate!(caps_ctrl_shift_ä, Key::new(ctrl, 'Ä'), None); // can't match without shifted key
-    validate!(caps_ctrl_shift_ä, Key::new(ctrl_shift, 'Ä'), None);
-    // With a shifted codepoint, we can match the alternative notation too.
-    let caps_ctrl_shift_ä = KeyEvent::new_with(ctrl_shift, 'ä', Some('Ä'), None);
-    validate!(caps_ctrl_shift_ä, Key::new(ctrl_shift, 'ä'), Some(exact)); // canonical notation
-    validate!(caps_ctrl_shift_ä, Key::new(ctrl, 'ä'), None);
-    validate!(caps_ctrl_shift_ä, Key::new(ctrl, 'Ä'), Some(modulo_shift)); // matched via shifted key
-    validate!(caps_ctrl_shift_ä, Key::new(ctrl_shift, 'Ä'), None);
-
-    let ctrl_ц = KeyEvent::new_with(ctrl, 'ц', None, Some('w'));
-    let ctrl_shift_ц = KeyEvent::new_with(ctrl_shift, 'ц', Some('Ц'), Some('w'));
-    validate!(ctrl_ц, Key::new(ctrl, 'ц'), Some(exact));
-    validate!(ctrl_ц, Key::new(ctrl, 'w'), Some(base_layout));
-    validate!(ctrl_ц, Key::new(ctrl_shift, 'ц'), None);
-    validate!(ctrl_ц, Key::new(ctrl_shift, 'w'), None);
-    validate!(
-        ctrl_shift_ц,
-        Key::new(ctrl, 'W'),
-        Some(base_layout_modulo_shift)
-    );
-    validate!(ctrl_shift_ц, Key::new(ctrl, 'w'), None);
-
-    // Note that "bind ctrl-Ц" will win over "bind ctrl-shift-w".
-    // This is because we consider shift transformation to be less magic than base-key
-    // transformation.
-    validate!(ctrl_shift_ц, Key::new(ctrl, 'Ц'), Some(modulo_shift));
-    validate!(ctrl_shift_ц, Key::new(ctrl_shift, 'w'), Some(base_layout));
-}
-
 /// Represents an event on the character input stream.
 #[derive(Debug, Clone)]
 pub enum CharEventType {
@@ -402,16 +318,14 @@ pub enum ImplicitEvent {
     /// An event was handled internally, or an interrupt was received. Check to see if the reader
     /// loop should exit.
     CheckExit,
-    /// A blocking terminal query was interrupterd with ctrl-c.
-    QueryInterrupted,
     /// Our terminal window gained focus.
     FocusIn,
     /// Our terminal window lost focus.
     FocusOut,
-    /// Request to disable mouse tracking.
-    DisableMouseTracking,
     /// Mouse left click.
     MouseLeft(ViewportPosition),
+    /// Window height changed.
+    WindowHeight,
 }
 
 #[derive(Debug, Clone)]
@@ -424,6 +338,8 @@ pub enum QueryResponse {
 pub enum QueryResultEvent {
     Response(QueryResponse),
     Timeout,
+    /// Canceled with ctrl-c.
+    Interrupted,
 }
 
 #[derive(Debug, Clone)]
@@ -560,14 +476,13 @@ fn readb(in_fd: RawFd) -> Option<u8> {
     Some(c)
 }
 
-fn next_input_event(in_fd: RawFd, timeout: Timeout) -> InputEventTrigger {
+fn next_input_event(in_fd: RawFd, ioport_fd: RawFd, timeout: Timeout) -> InputEventTrigger {
     let mut fdset = FdReadableSet::new();
     loop {
         fdset.clear();
         fdset.add(in_fd);
 
-        // Add the completion ioport.
-        let ioport_fd = iothread_port();
+        // Add the completion ioport (possibly -1 - a no-op).
         fdset.add(ioport_fd);
 
         // Get the uvar notifier fd (possibly none).
@@ -681,7 +596,7 @@ pub fn update_wait_on_escape_ms(vars: &EnvStack) {
         _ => {
             eprintf!(
                 concat!(
-                    "ignoring fish_escape_delay_ms: value '%ls' ",
+                    "ignoring fish_escape_delay_ms: value '%s' ",
                     "is not an integer or is < 10 or >= 5000 ms\n"
                 ),
                 fish_escape_delay_ms
@@ -706,7 +621,7 @@ pub fn update_wait_on_sequence_key_ms(vars: &EnvStack) {
         _ => {
             eprintf!(
                 concat!(
-                    "ignoring fish_sequence_key_delay_ms: value '%ls' ",
+                    "ignoring fish_sequence_key_delay_ms: value '%s' ",
                     "is not an integer or is < 10 or >= 5000 ms\n"
                 ),
                 sequence_key_time_ms
@@ -782,20 +697,23 @@ impl InputData {
 }
 
 #[derive(Clone, Eq, PartialEq)]
-pub enum CursorPositionQueryKind {
-    MouseLeft(ViewportPosition),
-    ScrollbackPush,
+pub enum CursorPositionQueryReason {
+    NewPrompt,
+    WindowHeightChange,
 }
 
 #[derive(Clone, Eq, PartialEq)]
 pub struct CursorPositionQuery {
-    pub kind: CursorPositionQueryKind,
+    pub reason: CursorPositionQueryReason,
     pub result: Option<ViewportPosition>,
 }
 
 impl CursorPositionQuery {
-    pub fn new(kind: CursorPositionQueryKind) -> Self {
-        Self { kind, result: None }
+    pub fn new(reason: CursorPositionQueryReason) -> Self {
+        Self {
+            reason,
+            result: None,
+        }
     }
 }
 
@@ -813,8 +731,7 @@ pub trait InputEventQueuer {
         if self.is_blocked_querying() {
             use ImplicitEvent::*;
             match self.get_input_data().queue.front()? {
-                CharEvent::QueryResult(_)
-                | CharEvent::Implicit(CheckExit | Eof | QueryInterrupted) => {}
+                CharEvent::QueryResult(_) | CharEvent::Implicit(CheckExit | Eof) => {}
                 CharEvent::Key(_)
                 | CharEvent::Readline(_)
                 | CharEvent::Command(_)
@@ -826,9 +743,7 @@ pub trait InputEventQueuer {
         self.get_input_data_mut().queue.pop_front()
     }
 
-    /// Function used by [`readch`](Self::readch) to read bytes from stdin until enough bytes have been read to
-    /// convert them to a wchar_t. Conversion is done using mbrtowc. If a character has previously
-    /// been read and then 'unread' using \c input_common_unreadch, that character is returned.
+    /// Read the next event, such as a UTF-8-encoded codepoint.
     fn readch(&mut self) -> CharEvent {
         loop {
             // Do we have something enqueued already?
@@ -846,6 +761,7 @@ pub trait InputEventQueuer {
 
             match next_input_event(
                 self.get_in_fd(),
+                self.get_ioport_fd(),
                 if self.is_blocked_querying() {
                     Timeout::Duration(self.get_input_data().blocking_query_timeout.unwrap())
                 } else {
@@ -871,7 +787,7 @@ pub trait InputEventQueuer {
                 InputEventTrigger::Byte(read_byte) => {
                     let mut have_escape_prefix = false;
                     let mut buffer = vec![read_byte];
-                    let key_with_escape = if read_byte == 0x1b {
+                    let mut key = if read_byte == 0x1b {
                         self.parse_escape_sequence(&mut buffer, &mut have_escape_prefix)
                     } else {
                         canonicalize_control_char(read_byte).map(KeyEvent::from)
@@ -883,47 +799,39 @@ pub trait InputEventQueuer {
                         continue;
                     }
                     let mut seq = WString::new();
-                    let mut key = key_with_escape;
                     if key.is_some_and(|key| key.key == Key::from_raw(key::Invalid)) {
                         continue;
                     }
-                    assert!(key.map_or(true, |key| key.codepoint != key::Invalid));
-                    let mut consumed = 0;
-                    let mut state = zero_mbstate();
-                    let mut i = 0;
+                    assert!(key.is_none_or(|key| key.codepoint != key::Invalid));
+                    // At this point, the bytes in `buffer` should be parsed as a UTF-8 sequence,
+                    // or, if they are not valid UTF-8, ignored. On incomplete sequences, another
+                    // byte is read and decoding is tried again in the next iteration.
                     let ok = loop {
-                        if i == buffer.len() {
-                            buffer.push(
-                                match next_input_event(self.get_in_fd(), Timeout::Forever) {
-                                    InputEventTrigger::Byte(b) => b,
-                                    _ => 0,
-                                },
-                            );
-                        }
-                        match decode_input_byte(
-                            &mut seq,
-                            InvalidPolicy::Error,
-                            &mut state,
-                            &buffer[..i + 1],
-                            &mut consumed,
-                        ) {
-                            DecodeState::Incomplete => (),
+                        match decode_one_codepoint_utf8(&mut seq, InvalidPolicy::Error, &buffer) {
+                            DecodeState::Incomplete => {
+                                buffer.push(
+                                    match next_input_event(
+                                        self.get_in_fd(),
+                                        self.get_ioport_fd(),
+                                        Timeout::Forever,
+                                    ) {
+                                        InputEventTrigger::Byte(b) => b,
+                                        _ => 0,
+                                    },
+                                );
+                            }
                             DecodeState::Complete => {
-                                if have_escape_prefix && i != 0 {
-                                    have_escape_prefix = false;
+                                if have_escape_prefix {
                                     let c = seq.as_char_slice().last().unwrap();
                                     key = Some(KeyEvent::from(alt(*c)));
                                 }
-                                if i + 1 == buffer.len() {
-                                    break true;
-                                }
+                                break true;
                             }
                             DecodeState::Error => {
                                 self.push_front(CharEvent::from_check_exit());
                                 break false;
                             }
                         }
-                        i += 1;
                     };
                     if !ok {
                         continue;
@@ -965,7 +873,7 @@ pub trait InputEventQueuer {
                             let ok = stop_query(self.blocking_query());
                             assert!(ok);
                             self.get_input_data_mut().queue.clear();
-                            self.push_front(CharEvent::Implicit(ImplicitEvent::QueryInterrupted));
+                            self.push_front(CharEvent::QueryResult(QueryResultEvent::Interrupted));
                         }
                         continue;
                     }
@@ -1143,7 +1051,7 @@ pub trait InputEventQueuer {
             b'F' => masked_key(key::End),  // PC/xterm style
             b'H' => masked_key(key::Home), // PC/xterm style
             b'M' | b'm' => {
-                self.disable_mouse_tracking();
+                FLOG!(reader, "mouse event");
                 // Generic X10 or modified VT200 sequence, or extended (SGR/1006) mouse
                 // reporting mode, with semicolon-separated parameters for button code, Px,
                 // and Py, ending with 'M' for button press or 'm' for button release.
@@ -1183,14 +1091,14 @@ pub trait InputEventQueuer {
                 return None;
             }
             b't' => {
-                self.disable_mouse_tracking();
+                FLOG!(reader, "mouse event");
                 // VT200 button released in mouse highlighting mode at valid text location. 5 chars.
                 let _ = next_char(self);
                 let _ = next_char(self);
                 return None;
             }
             b'T' => {
-                self.disable_mouse_tracking();
+                FLOG!(reader, "mouse event");
                 // VT200 button released in mouse highlighting mode past end-of-line. 9 characters.
                 for _ in 0..6 {
                     let _ = next_char(self);
@@ -1343,22 +1251,12 @@ pub trait InputEventQueuer {
         Some(key)
     }
 
-    fn disable_mouse_tracking(&mut self) {
-        // fish recognizes but does not actually support mouse reporting. We never turn it on, and
-        // it's only ever enabled if a program we spawned enabled it and crashed or forgot to turn
-        // it off before exiting. We turn it off here to avoid wasting resources.
-        FLOG!(reader, "Disabling mouse tracking");
-
-        // We shouldn't directly manipulate stdout from here, so we ask the reader to do it.
-        self.push_front(CharEvent::Implicit(ImplicitEvent::DisableMouseTracking));
-    }
-
     fn parse_ss3(&mut self, buffer: &mut Vec<u8>) -> Option<KeyEvent> {
         let mut raw_mask = 0;
         let Some(mut code) = self.try_readb(buffer) else {
             return Some(KeyEvent::from(alt('O')));
         };
-        while (b'0'..=b'9').contains(&code) {
+        while code.is_ascii_digit() {
             raw_mask = raw_mask * 10 + u32::from(code - b'0');
             code = self.try_readb(buffer).unwrap_or(0xff);
         }
@@ -1421,7 +1319,7 @@ pub trait InputEventQueuer {
             return None;
         }
         XTVERSION.get_or_init(|| {
-            let xtversion = str2wcstring(&buffer[4..buffer.len()]);
+            let xtversion = bytes2wcstring(&buffer[4..buffer.len()]);
             FLOG!(
                 reader,
                 format!("Received XTVERSION response: {}", xtversion)
@@ -1460,7 +1358,7 @@ pub trait InputEventQueuer {
                 reader,
                 format!(
                     "Received XTGETTCAP failure response: {}",
-                    str2wcstring(&parse_hex(buffer)?),
+                    bytes2wcstring(&parse_hex(buffer)?),
                 )
             );
             return None;
@@ -1474,14 +1372,14 @@ pub trait InputEventQueuer {
                 reader,
                 format!(
                     "Received XTGETTCAP response: {}={:?}",
-                    str2wcstring(&key),
-                    str2wcstring(&value)
+                    bytes2wcstring(&key),
+                    bytes2wcstring(&value)
                 )
             );
         } else {
             FLOG!(
                 reader,
-                format!("Received XTGETTCAP response: {}", str2wcstring(&key))
+                format!("Received XTGETTCAP response: {}", bytes2wcstring(&key))
             );
         }
         if key == SCROLL_CONTENT_UP_TERMINFO_CODE.as_bytes() {
@@ -1522,6 +1420,11 @@ pub trait InputEventQueuer {
         self.get_input_data().in_fd
     }
 
+    /// Return the fd of the IO port, or -1 if none.
+    fn get_ioport_fd(&self) -> RawFd {
+        -1
+    }
+
     /// Return the input data. This is to be implemented by the concrete type.
     fn get_input_data(&self) -> &InputData;
     fn get_input_data_mut(&mut self) -> &mut InputData;
@@ -1648,7 +1551,7 @@ pub trait InputEventQueuer {
                     "Received interrupt, giving up on waiting for terminal response"
                 );
                 self.get_input_data_mut().queue.clear();
-                self.push_front(CharEvent::Implicit(ImplicitEvent::QueryInterrupted));
+                self.push_front(CharEvent::QueryResult(QueryResultEvent::Interrupted));
             } else {
                 self.push_front(interrupt_evt);
             }
@@ -1686,65 +1589,37 @@ pub(crate) enum InvalidPolicy {
     Passthrough,
 }
 
-pub(crate) fn decode_input_byte(
+pub(crate) fn decode_one_codepoint_utf8(
     out_seq: &mut WString,
     invalid_policy: InvalidPolicy,
-    state: &mut mbstate_t,
     buffer: &[u8],
-    consumed: &mut usize,
 ) -> DecodeState {
     use DecodeState::*;
-    let mut res: char = '\0';
-    let read_byte = *buffer.last().unwrap();
-    if crate::libc::MB_CUR_MAX() == 1 {
-        // single-byte locale, all values are legal
-        // FIXME: this looks wrong, this falsely assumes that
-        // the single-byte locale is compatible with Unicode upper-ASCII.
-        res = read_byte.into();
-        out_seq.push(res);
-        return Complete;
-    }
-    let mut invalid = |out_seq: &mut WString, log_error: fn()| match invalid_policy {
-        InvalidPolicy::Error => {
-            (log_error)();
-            Error
-        }
-        InvalidPolicy::Passthrough => {
-            for &b in &buffer[*consumed..] {
-                out_seq.push(encode_byte_to_char(b));
+    match std::str::from_utf8(buffer) {
+        Ok(parsed_str) => {
+            for c in parsed_str.chars() {
+                if !fish_reserved_codepoint(c) {
+                    out_seq.push(c);
+                }
             }
-            *consumed = buffer.len();
             Complete
         }
-    };
-    let mut codepoint = u32::from(res);
-    match unsafe {
-        mbrtowc(
-            std::ptr::addr_of_mut!(codepoint),
-            std::ptr::addr_of!(read_byte).cast(),
-            1,
-            state,
-        )
-    } as isize
-    {
-        -1 => {
-            return invalid(out_seq, || FLOG!(reader, "Illegal input encoding"));
-        }
-        -2 => {
-            // Sequence not yet complete.
-            return Incomplete;
-        }
-        _ => (),
+        Err(e) => match e.error_len() {
+            Some(_) => match invalid_policy {
+                InvalidPolicy::Error => {
+                    FLOG!(reader, "Illegal input encoding");
+                    Error
+                }
+                InvalidPolicy::Passthrough => {
+                    for &b in buffer {
+                        out_seq.push(encode_byte_to_char(b));
+                    }
+                    Complete
+                }
+            },
+            None => Incomplete,
+        },
     }
-    if let Some(res) = char::from_u32(codepoint) {
-        // Sequence complete.
-        if !fish_reserved_codepoint(res) {
-            *consumed += 1;
-            out_seq.push(res);
-            return Complete;
-        }
-    }
-    invalid(out_seq, || FLOG!(reader, "Illegal codepoint"))
 }
 
 pub(crate) fn stop_query(mut query: RefMut<'_, Option<TerminalQuery>>) -> bool {
@@ -1820,7 +1695,156 @@ fn parse_hex(hex: &[u8]) -> Option<Vec<u8>> {
     Some(result)
 }
 
-#[test]
-fn test_parse_hex() {
-    assert_eq!(parse_hex(b"3d"), Some(vec![61]));
+#[cfg(test)]
+mod tests {
+    use super::{
+        CharEvent, InputEventQueue, InputEventQueuer, KeyEvent, KeyMatchQuality, ReadlineCmd,
+        match_key_event_to_key, parse_hex,
+    };
+    use crate::key::{Key, Modifiers};
+
+    #[test]
+    fn test_match_key_event_to_key() {
+        macro_rules! validate {
+            ($evt:expr, $key:expr, $expected:expr) => {
+                assert_eq!(match_key_event_to_key(&$evt, &$key), $expected);
+            };
+        }
+
+        let none = Modifiers::default();
+        let shift = Modifiers::SHIFT;
+        let ctrl = Modifiers::CTRL;
+        let ctrl_shift = Modifiers {
+            ctrl: true,
+            shift: true,
+            ..Default::default()
+        };
+
+        let exact = KeyMatchQuality::Exact;
+        let modulo_shift = KeyMatchQuality::ModuloShift;
+        let base_layout = KeyMatchQuality::BaseLayout;
+        let base_layout_modulo_shift = KeyMatchQuality::BaseLayoutModuloShift;
+
+        validate!(KeyEvent::new(none, 'a'), Key::new(none, 'a'), Some(exact));
+        validate!(KeyEvent::new(none, 'a'), Key::new(none, 'A'), None);
+        validate!(KeyEvent::new(shift, 'a'), Key::new(shift, 'a'), Some(exact));
+        validate!(KeyEvent::new(shift, 'a'), Key::new(none, 'A'), None);
+        validate!(KeyEvent::new(shift, 'ä'), Key::new(none, 'Ä'), None);
+        // For historical reasons we canonicalize notation for ASCII keys like "shift-a" to "A",
+        // but not "shift-a" events - those should send a shifted key.
+        validate!(
+            KeyEvent::new(none, 'A'),
+            Key::new(shift, 'a'),
+            Some(modulo_shift)
+        );
+        validate!(KeyEvent::new(none, 'A'), Key::new(shift, 'A'), None);
+        validate!(KeyEvent::new(none, 'Ä'), Key::new(none, 'Ä'), Some(exact));
+        validate!(KeyEvent::new(none, 'Ä'), Key::new(shift, 'ä'), None);
+
+        // FYI: for codepoints that are not letters with uppercase/lowercase versions, we use
+        // the shifted key in the canonical notation, because the unshifted one may depend on the
+        // keyboard layout.
+        let ctrl_shift_equals = KeyEvent::new_with(ctrl_shift, '=', Some('+'), None);
+        validate!(ctrl_shift_equals, Key::new(ctrl_shift, '='), Some(exact));
+        validate!(ctrl_shift_equals, Key::new(ctrl, '+'), Some(modulo_shift)); // canonical notation
+        validate!(ctrl_shift_equals, Key::new(ctrl_shift, '+'), None);
+        validate!(ctrl_shift_equals, Key::new(ctrl, '='), None);
+
+        // A event like capslock-shift-ä may or may not include a shifted codepoint.
+        //
+        // Without a shifted codepoint, we cannot easily match ctrl-Ä.
+        let caps_ctrl_shift_ä = KeyEvent::new(ctrl_shift, 'ä');
+        validate!(caps_ctrl_shift_ä, Key::new(ctrl_shift, 'ä'), Some(exact)); // canonical notation
+        validate!(caps_ctrl_shift_ä, Key::new(ctrl, 'ä'), None);
+        validate!(caps_ctrl_shift_ä, Key::new(ctrl, 'Ä'), None); // can't match without shifted key
+        validate!(caps_ctrl_shift_ä, Key::new(ctrl_shift, 'Ä'), None);
+        // With a shifted codepoint, we can match the alternative notation too.
+        let caps_ctrl_shift_ä = KeyEvent::new_with(ctrl_shift, 'ä', Some('Ä'), None);
+        validate!(caps_ctrl_shift_ä, Key::new(ctrl_shift, 'ä'), Some(exact)); // canonical notation
+        validate!(caps_ctrl_shift_ä, Key::new(ctrl, 'ä'), None);
+        validate!(caps_ctrl_shift_ä, Key::new(ctrl, 'Ä'), Some(modulo_shift)); // matched via shifted key
+        validate!(caps_ctrl_shift_ä, Key::new(ctrl_shift, 'Ä'), None);
+
+        let ctrl_ц = KeyEvent::new_with(ctrl, 'ц', None, Some('w'));
+        let ctrl_shift_ц = KeyEvent::new_with(ctrl_shift, 'ц', Some('Ц'), Some('w'));
+        validate!(ctrl_ц, Key::new(ctrl, 'ц'), Some(exact));
+        validate!(ctrl_ц, Key::new(ctrl, 'w'), Some(base_layout));
+        validate!(ctrl_ц, Key::new(ctrl_shift, 'ц'), None);
+        validate!(ctrl_ц, Key::new(ctrl_shift, 'w'), None);
+        validate!(
+            ctrl_shift_ц,
+            Key::new(ctrl, 'W'),
+            Some(base_layout_modulo_shift)
+        );
+        validate!(ctrl_shift_ц, Key::new(ctrl, 'w'), None);
+
+        // Note that "bind ctrl-Ц" will win over "bind ctrl-shift-w".
+        // This is because we consider shift transformation to be less magic than base-key
+        // transformation.
+        validate!(ctrl_shift_ц, Key::new(ctrl, 'Ц'), Some(modulo_shift));
+        validate!(ctrl_shift_ц, Key::new(ctrl_shift, 'w'), Some(base_layout));
+    }
+
+    #[test]
+    fn test_push_front_back() {
+        let mut queue = InputEventQueue::new(0, None);
+        queue.push_front(CharEvent::from_char('a'));
+        queue.push_front(CharEvent::from_char('b'));
+        queue.push_back(CharEvent::from_char('c'));
+        queue.push_back(CharEvent::from_char('d'));
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'b');
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'a');
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'c');
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'd');
+        assert!(queue.try_pop().is_none());
+    }
+
+    #[test]
+    fn test_promote_interruptions_to_front() {
+        let mut queue = InputEventQueue::new(0, None);
+        queue.push_back(CharEvent::from_char('a'));
+        queue.push_back(CharEvent::from_char('b'));
+        queue.push_back(CharEvent::from_readline(ReadlineCmd::Undo));
+        queue.push_back(CharEvent::from_readline(ReadlineCmd::Redo));
+        queue.push_back(CharEvent::from_char('c'));
+        queue.push_back(CharEvent::from_char('d'));
+        queue.promote_interruptions_to_front();
+
+        assert_eq!(queue.try_pop().unwrap().get_readline(), ReadlineCmd::Undo);
+        assert_eq!(queue.try_pop().unwrap().get_readline(), ReadlineCmd::Redo);
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'a');
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'b');
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'c');
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'd');
+        assert!(!queue.has_lookahead());
+
+        queue.push_back(CharEvent::from_char('e'));
+        queue.promote_interruptions_to_front();
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'e');
+        assert!(!queue.has_lookahead());
+    }
+
+    #[test]
+    fn test_insert_front() {
+        let mut queue = InputEventQueue::new(0, None);
+        queue.push_back(CharEvent::from_char('a'));
+        queue.push_back(CharEvent::from_char('b'));
+
+        let events = vec![
+            CharEvent::from_char('A'),
+            CharEvent::from_char('B'),
+            CharEvent::from_char('C'),
+        ];
+        queue.insert_front(events);
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'A');
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'B');
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'C');
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'a');
+        assert_eq!(queue.try_pop().unwrap().get_char(), 'b');
+    }
+
+    #[test]
+    fn test_parse_hex() {
+        assert_eq!(parse_hex(b"3d"), Some(vec![61]));
+    }
 }
